// =============================================================================
//
//   Simple Bumpmapping - glPrimitives.pas
//
// =============================================================================
//   Copyright  2003-2004 by Sascha Willems - http://www.delphigl.de
//   visit the Delphi OpenGL Community - http://www.delphigl.com
// =============================================================================
//   Contents of this file are subject to the GNU Public License (GPL) which can
//   be obtained here : http://opensource.org/licenses/gpl-license.php
// =============================================================================
//   This should have been a unit to create several types of primitives, but
//   right now it only supports a simple torus.
//   To inlcude as less units as necessary, I also put some routines from
//   my gl-mathlibrary in here.
// =============================================================================

unit glPrimitives;

interface

uses
 dglOpenGL;

// Since I don't want to annoy you with my complete math-library, I copied the
// used types and functions to this unit...
type
 TglVertex3f = record
   x,y,z  : TglFloat;
  end;
 TglVertexTBN = packed record
   s,t               : Single;
   Normal            : TglVertex3f;
   sTangent          : TglVertex3f;
   tTangent          : TglVertex3f; // Also called BiNormal
  end;

type
 TglPrimitiveType = (TglTorus);
 TglPrimitive = class(TObject)
   private
    procedure GenerateTorus(pPrecision : Integer);
   public
    Vertex            : array of TglVertex3f;
    Tangent           : array of TglVertexTBN;
    Index             : array of Integer;
    TangentSpaceLight : array of TglVertex3f;
    constructor Create(pType : TglPrimitiveType);
    destructor Destroy; override;
    procedure Generate(pType : TglPrimitiveType);
  end;

// Some functions from my mathlib...
function glVertex(fX, fY, fZ : TGLFloat) : TglVertex3f;
function glAddVector(fV1, fV2 : TglVertex3f) : TglVertex3f;
procedure glNormalizeVector(var v : TglVertex3f);
function glScaleVector(fV1, fV2 : TglVertex3f) : TglVertex3f;
function glMult3f(v: TglVertex3f;m : TGLMatrixf4) : TglVertex3f;
function glSubtractVector(fV1, fV2 : TglVertex3f) : TglVertex3f;
function glDotProduct(pV1, pV2 : TglVertex3f) : TGLFloat;

implementation

// =============================================================================
// =============================================================================
// Some functions from my mathlib...
// =============================================================================

// =============================================================================
//  glVertex
// =============================================================================
function glVertex(fX, fY, fZ : TGLFloat) : TglVertex3f;
begin
Result.x := fX;
Result.y := fY;
Result.z := fZ;
end;

// =============================================================================
//  glAddVector
// =============================================================================
function glAddVector(fV1, fV2 : TglVertex3f) : TglVertex3f;
begin
Result.x := fV1.x+fV2.x;
Result.y := fV1.y+fV2.y;
Result.z := fV1.z+fV2.z;
end;

// =============================================================================
//  glGetRotatedZ
// =============================================================================
function glGetRotatedZ(v : TglVertex3f;fAngle : Single) : TglVertex3f;
var
 SinAngle,CosAngle : Single;
begin
if fAngle = 0 then
 Result := v;
SinAngle := Sin(PI*fAngle/180);
CosAngle := Cos(PI*fAngle/180);
Result   := glVertex(v.x*CosAngle-v.y*SinAngle, v.x*SinAngle+v.y*CosAngle, v.z);
end;

// =============================================================================
//  glGetRotatedY
// =============================================================================
function glGetRotatedY(v : TglVertex3f;fAngle : Single) : TglVertex3f;
var
 SinAngle,CosAngle : Single;
begin
if fAngle = 0 then
 Result := v;
SinAngle := Sin(PI*fAngle/180);
CosAngle := Cos(PI*fAngle/180);
Result   := glVertex(v.x*CosAngle+v.z*SinAngle, v.y, -v.x*SinAngle+v.z*CosAngle);
end;

// =============================================================================
//  glCrossProduct
// =============================================================================
function glCrossProduct(pV1, pV2 : TglVertex3f) : TglVertex3f;
begin
with Result do
 begin
 x := (pV1.y*pV2.z) - (pV1.z*pV2.y);
 y := (pV1.z*pV2.x) - (pV1.x*pV2.z);
 z := (pV1.x*pV2.y) - (pV1.y*pV2.x);
 end;
end;

// =============================================================================
//  glNormalizeVector
// =============================================================================
procedure glNormalizeVector(var v : TglVertex3f);
var
 l : Single;
begin
l := Sqrt(v.x*v.x + v.y*v.y + v.z*v.z);
if l <> 0 then
 with v do
  begin
  x := x/l;
  y := y/l;
  z := z/l;
  end;
end;

// =============================================================================
//  glScaleVector
// =============================================================================
function glScaleVector(fV1, fV2 : TglVertex3f) : TglVertex3f;
begin
Result.x := fV1.x*fV2.x;
Result.y := fV1.y*fV2.y;
Result.z := fV1.z*fV2.z;
end;

// =============================================================================
//  glMult3f
// =============================================================================
function glMult3f(v: TglVertex3f;m : TGLMatrixf4) : TglVertex3f;
begin
Result.x := (v.x*m[0,0])+(v.y*m[0,1])+(v.z*m[0,2])+(1*m[0,3]);
Result.y := (v.x*m[1,0])+(v.y*m[1,1])+(v.z*m[1,2])+(1*m[1,3]);
Result.z := (v.x*m[2,0])+(v.y*m[2,1])+(v.z*m[2,2])+(1*m[2,3]);
end;

// =============================================================================
//  glSubtractVector
// =============================================================================
function glSubtractVector(fV1, fV2 : TglVertex3f) : TglVertex3f;
begin
Result.x := fV1.x-fV2.x;
Result.y := fV1.y-fV2.y;
Result.z := fV1.z-fV2.z;
end;

// =============================================================================
//  glDotProduct
// =============================================================================
function glDotProduct(pV1, pV2 : TglVertex3f) : TGLFloat;
begin
Result := (pV1.x*pV2.x) + (pV1.y*pV2.y) + (pV1.z*pV2.z);
end;

// =============================================================================
// =============================================================================


constructor TglPrimitive.Create(pType : TglPrimitiveType);
begin
inherited Create;
Generate(pType);
end;

destructor TglPrimitive.Destroy;
begin
SetLength(Vertex,0);
SetLength(Index,0);
inherited;
end;

procedure TglPrimitive.Generate(pType : TglPrimitiveType);
begin
case pType of
 TglTorus : GenerateTorus(48);
end;
end;

procedure TglPrimitive.GenerateTorus(pPrecision : Integer);
var
 i,Ring : Integer;
begin
SetLength(Vertex, (pPrecision+1)*(pPrecision+1));
SetLength(Tangent, (pPrecision+1)*(pPrecision+1));
SetLength(TangentSpaceLight, (pPrecision+1)*(pPrecision+1));
SetLength(Index, 2*pPrecision*pPrecision*3-1);
for i := 0 to pPrecision do
 begin
 Vertex[i]           := glAddVector(glGetRotatedZ(glVertex(1.5,0,0),i*360/pPrecision),glVertex(4,0,0));
 Tangent[i].s        := 0;
 Tangent[i].t        := i/pPrecision*1.5;
 Tangent[i].sTangent := glVertex(0,0,-1);
 Tangent[i].tTangent := glGetRotatedZ(glVertex(0,-1,0),i*360/pPrecision);
 Tangent[i].Normal   := glCrossProduct(Tangent[i].tTangent, Tangent[i].sTangent);
 end;
for Ring := 1 to pPrecision do
 for i := 0 to pPrecision do
  begin
  Vertex[Ring*(pPrecision+1)+i]           := glGetRotatedY(Vertex[i],Ring*360/pPrecision);
  Tangent[Ring*(pPrecision+1)+i].s        := 2*Ring/pPrecision*1.5;
  Tangent[Ring*(pPrecision+1)+i].t        := Tangent[i].t;
  Tangent[Ring*(pPrecision+1)+i].sTangent := glGetRotatedY(Tangent[i].sTangent,Ring*360/pPrecision);
  Tangent[Ring*(pPrecision+1)+i].tTangent := glGetRotatedY(Tangent[i].tTangent,Ring*360/pPrecision);
  Tangent[Ring*(pPrecision+1)+i].Normal   := glGetRotatedY(Tangent[i].Normal,Ring*360/pPrecision);
  end;
for Ring := 0 to pPrecision-1 do
 for i := 0 to pPrecision-1 do
  begin
  Index[((Ring*pPrecision+i)*2)*3+0]   := Ring*(pPrecision+1)+i;
  Index[((Ring*pPrecision+i)*2)*3+1]   := (Ring+1)*(pPrecision+1)+i;
  Index[((Ring*pPrecision+i)*2)*3+2]   := Ring*(pPrecision+1)+i+1;
  Index[((Ring*pPrecision+i)*2+1)*3+0] := Ring*(pPrecision+1)+i+1;
  Index[((Ring*pPrecision+i)*2+1)*3+1] := (Ring+1)*(pPrecision+1)+i;
  Index[((Ring*pPrecision+i)*2+1)*3+2] := (Ring+1)*(pPrecision+1)+i+1;
  end;
exit;
end;


end.
